#ifndef _MODEL_H
#define _MODEL_H

#include <string>
#include <iostream>
#include <cassert>

#ifdef __APPLE__
#include <GLExtensionWrangler/glew.h>
#include <OpenGL/gl.h>
#else
#include <GL/glew.h>
#include <GL/gl.h>
#endif

#include "vertexarray.h"
#include "mathvector.h"
#include "aabb.h"

///loading data into the mesh vertexarray is implemented by derived classes
class MODEL
{
private:
	bool generatedlistid;
	int listid;
	bool generatedmetrics;
	std::string diffuse_texture;
	std::string normal_texture;
	
	//metrics
	float radius;
	float radiusxz;
	MATHVECTOR <float, 3> bboxmin;
	MATHVECTOR <float, 3> bboxmax;
	
	void RequireMetrics() const
	{
		//Mesh metrics need to be generated before they can be queried
		assert(generatedmetrics);
	}
	
	void RequireListID() const
	{
		//Mesh id needs to be generated
		assert(generatedlistid);
	}
	
	void ClearListID();
	void ClearMetrics() {generatedmetrics = false;}
	
protected:
	///to be filled by the derived classes
	VERTEXARRAY mesh;
	std::string modelpath;
	std::string modelname;

public:
	MODEL() : generatedlistid(false),generatedmetrics(false),radius(0),radiusxz(0) {}
	virtual ~MODEL() {Clear();}
	
	void SetDiffuseTexture(const std::string & newdiff) {diffuse_texture = newdiff;}
	void SetNormalTexture(const std::string & newnorm) {normal_texture = newnorm;}
	const std::string GetDiffuseTexture() const {return diffuse_texture;}
	const std::string GetNormalTexture() const {return normal_texture;}
	
	void GenerateListID(std::ostream & error_output);
	void GenerateMeshMetrics();
	void ClearMeshData() {mesh.Clear();}
	
	int GetListID() const {RequireListID(); return listid;}
	
	//"metrics"
	float GetRadius() const {RequireMetrics();return radius + 0.5f;}
	float GetRadiusXZ() const {RequireMetrics();return radiusxz;}
	MATHVECTOR <float, 3> GetCenter() {return (bboxmax + bboxmin) * 0.5;}
	
	bool HaveMeshData() const {return (mesh.GetNumFaces() > 0);}
	bool HaveMeshMetrics() const {return generatedmetrics;}
	bool HaveListID() const {return generatedlistid;}
	
	void Clear() {ClearMeshData();ClearListID();ClearMetrics();}
	
	const VERTEXARRAY & GetVertexArray() {return mesh;}
	
	bool Loaded() {return (mesh.GetNumFaces() > 0);}
	
	AABB <float> GetAABB()
	{
		assert(generatedmetrics);
		AABB <float> output;
		output.SetFromCorners(bboxmin, bboxmax);
		return output;
	}
};

#endif


