#include "replay.h"
#include "unittest.h"

#include <iostream>
using std::endl;

#include <string>
using std::string;

#include <sstream>
using std::stringstream;

void REPLAY::Save(std::ostream & outstream)
{
	//write the file format version data manually.  if the serialization functions were used, a variable length string would be written instead, which isns't exactly what we want
	version_info.Save(outstream);
	
	SERIALIZER_BIN_OUTPUT serialize_output(outstream);
	serialize_output.SetPointerTracking(false);
	serialize_output.StartSerialize("replay", *this);
}

bool REPLAY::Load(std::istream & instream, std::ostream & error_output)
{
	REPLAYVERSIONING stream_version;
	stream_version.Load(instream);
	
	if (!(stream_version == version_info))
	{
		error_output << "Stream version " << stream_version.format_version << "/" << stream_version.inputs_supported << " does not match expected version " << version_info.format_version << "/" << version_info.inputs_supported << endl;
		return false;
	}
	
	SERIALIZER_BIN_INPUT serialize_input(instream);
	serialize_input.SetPointerTracking(false);
	serialize_input.StartSerialize("replay", *this);
	
	return true;
}

void REPLAY::StartRecording(const std::string & newreplayfilename, const std::string & newcartype, const std::string & newcarpaint, const std::string & carfilename, const std::string & trackname, ostream & error_log)
{
	frame = 0;
	replaymode = RECORDING;
	track = trackname;
	replayfilename = newreplayfilename;
	cartype = newcartype;
	carpaint = newcarpaint;
	inputbuffer.clear();
	inputbuffer.resize(CARINPUT::GAME_ONLY_INPUTS_START_HERE, 0);
	
	ifstream carfilestream(carfilename.c_str());
	if (!carfilestream)
		error_log << "Unable to open car file to put into replay: " << carfilename << endl;
	while (carfilestream)
	{
		string newline;
		std::getline(carfilestream, newline);
		carfile.append(newline);
		carfile.push_back('\n');
	}
}

bool REPLAY::StartPlaying(const std::string & newreplayfilename, std::ostream & error_output)
{
	frame = 0;
	replaymode = PLAYING;
	replayfilename = newreplayfilename;
	inputbuffer.clear();
	inputbuffer.resize(CARINPUT::GAME_ONLY_INPUTS_START_HERE, 0);
	
	ifstream replaystream(replayfilename.c_str(), ifstream::binary);
	if (!replaystream)
	{
		error_output << "Error loading replay file: " << replayfilename << endl;
		return false;
	}
	
	if (!Load(replaystream, error_output))
		return false;
	
	cur_inputframe = inputframes.begin();
	cur_stateframe = stateframes.begin();
	
	return true;
}

void REPLAY::StopRecording()
{
	replaymode = IDLE;
	totalframes = frame;
	
	if (!replayfilename.empty())
	{
		ofstream f(replayfilename.c_str());
		Save(f);
	}
}

void REPLAY::StopPlaying()
{
	replaymode = IDLE;
}

void REPLAY::RecordFrame(const std::vector <float> & inputs, CAR & car)
{
	if (!GetRecording())
		return;
	
	if (frame > 2000000000) //enforce a maximum recording time of about 92 days
	{
		StopRecording();
	}
	
	assert(inputbuffer.size() == CARINPUT::GAME_ONLY_INPUTS_START_HERE);
	assert((unsigned int) version_info.inputs_supported == CARINPUT::GAME_ONLY_INPUTS_START_HERE);
	
	
	//record inputs
	INPUTFRAME newinputframe(frame);
	
	for (unsigned int i = 0; i < CARINPUT::GAME_ONLY_INPUTS_START_HERE; i++)
	{
		if (inputs[i] != inputbuffer[i])
		{
			inputbuffer[i] = inputs[i];
			newinputframe.AddInput(i, inputs[i]);
		}
	}
	
	if (newinputframe.GetNumInputs() > 0)
		inputframes.push_back(newinputframe);
	
	
	//record state
	int framespersecond = 1.0/version_info.framerate;
	if (frame % framespersecond == 0) //once per second
	{
		stringstream statestream;
		SERIALIZER_BIN_OUTPUT serialize_output(statestream);
		serialize_output.SetPointerTracking(false);
		_STARTSERIALIZE_((&serialize_output),car);
		stateframes.push_back(STATEFRAME(frame));
		//stateframes.push_back(STATEFRAME(11189196)); //for debugging (in hex, 11189196 is AABBCC
		//cout << "Recording state size: " << statestream.str().length() << endl; //should be ~984
		stateframes.back().SetBinaryStateData(statestream.str());
	}
	
	frame++;
}

const std::vector <float> & REPLAY::PlayFrame(CAR & car)
{
	if (!GetPlaying())
		return inputbuffer;
	
	frame++;
	
	if (frame > totalframes)
	{
		StopPlaying();
		return inputbuffer;
	}
	
	assert(inputbuffer.size() == CARINPUT::GAME_ONLY_INPUTS_START_HERE);
	assert((unsigned int) version_info.inputs_supported == CARINPUT::GAME_ONLY_INPUTS_START_HERE);
	
	//fast forward through the inputframes until we're up to date
	while (cur_inputframe != inputframes.end() && cur_inputframe->GetFrame() <= frame)
	{
		ProcessPlayInputFrame(*cur_inputframe);
		cur_inputframe++;
	}
	
	//fast forward through the stateframes until we're up to date
	while (cur_stateframe != stateframes.end() && cur_stateframe->GetFrame() <= frame)
	{
		if (cur_stateframe->GetFrame() == frame) ProcessPlayStateFrame(*cur_stateframe, car);
		cur_stateframe++;
	}
	
	return inputbuffer;
}

void REPLAY::ProcessPlayInputFrame(const INPUTFRAME & frame)
{
	for (unsigned int i = 0; i < frame.GetNumInputs(); i++)
	{
		std::pair <int, float> input_pair = frame.GetInput(i);
		inputbuffer[input_pair.first] = input_pair.second;
	}
}

void REPLAY::ProcessPlayStateFrame(const STATEFRAME & frame, CAR & car)
{
	//cout << "Playing state frame..." << endl;
	stringstream statestream(frame.GetBinaryStateData());
	//cout << "Frame size: " << frame.GetBinaryStateData().length() << endl;
	SERIALIZER_BIN_INPUT serialize_input(statestream);
	serialize_input.SetPointerTracking(false);
	_STARTSERIALIZE_((&serialize_input),car);
	//cout << "Played state frame" << endl;
}

QT_TEST(replay_test)
{
	/*//basic version validity check
	{
		REPLAY replay(0.004);
		stringstream teststream;
		replay.Save(teststream);
		QT_CHECK(replay.Load(teststream, std::cerr));
	}*/
}
